/*
 * Copyright (c) 1999-2024 Hartmut Seichter
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#ifndef PW_CORE_AXISANGLE_HPP
#define PW_CORE_AXISANGLE_HPP

#include <pw/core/matrix.hpp>
#include <pw/core/vector.hpp>

namespace pw {

template <std::floating_point Scalar> struct axisangle final {

    using value_type = Scalar;
    using axis_type  = vector3<Scalar>;

    axis_type axis = axis_type::basis(2);
    Scalar angle{};

    constexpr static auto
    from_matrix(const matrix<Scalar, 3, 3>& m) noexcept -> axisangle {

        return {.axis = axis_type{m[2][1] - m[1][2], // x
                                  m[0][2] - m[2][0], // y
                                  m[1][0] - m[0][1]} // z
                            .normalized(),
                .angle =
                    std::acos((m[0][0] + m[1][1] + m[2][2] - 1) / Scalar{2})};
    }

    constexpr auto to_matrix() const noexcept -> matrix<Scalar, 3, 3> {

        const auto axis_n = axis.normalized(); // always normalize

        const auto cos_a   = std::cos(angle);
        const auto sin_a   = std::sin(angle);
        const auto cos_1_a = 1 - cos_a;

        const auto v1_a = axis_n.x() * axis_n.y() * cos_1_a;
        const auto v2_a = axis_n.z() * sin_a;

        const auto v1_b = axis_n.x() * axis_n.z() * cos_1_a;
        const auto v2_b = axis_n.y() * sin_a;

        const auto v1_c = axis_n.y() * axis_n.z() * cos_1_a;
        const auto v2_c = axis_n.x() * sin_a;

        return {
            pw::vector{cos_a + axis_n.x() * axis_n.x() * cos_1_a, // [0,0]
                       v1_a - v2_a,                               // [0,1]
                       v1_b + v2_b},                              // [0,2]
            pw::vector{v1_a + v2_a,                               // [1,0]
                       cos_a + axis_n.y() * axis_n.y() * cos_1_a, // [1,1]
                       v1_c - v2_c},                              // [1,2]
            pw::vector{v1_b - v2_b,                               // [2,0]
                       v1_c + v2_c,                               // [2,1]
                       cos_a + axis_n.z() * axis_n.z() * cos_1_a} // [2,2]
        };
    }
};

using axisanglef = axisangle<float>;
using axisangled = axisangle<double>;

} // namespace pw

#endif
