/*
 * Copyright (c) 1999-2024 Hartmut Seichter
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef PW_CORE_FRUSTUM_HPP
#define PW_CORE_FRUSTUM_HPP

#include <pw/core/globals.hpp>
#include <pw/core/math.hpp>

namespace pw {

template <std::floating_point Scalar> struct frustum final {
    Scalar left{-1}, right{1}, bottom{-1}, top{1}, z_near{-1}, z_far{1};

    static constexpr auto make_perspective_symmetric(Scalar fov_h_deg,
                                                     Scalar aspect_ratio,
                                                     Scalar z_near,
                                                     Scalar z_far) -> frustum {
        const auto tangent_half{
            std::tan(pw::deg_to_rad(fov_h_deg / Scalar{2}))};
        const auto top{tangent_half * z_near};
        const auto right{top * aspect_ratio};

        return {.left   = -right,
                .right  = right,
                .bottom = -top,
                .top    = top,
                .z_near = z_near,
                .z_far  = z_far};
    }

    static constexpr auto make_orthographic_symmetric(Scalar scale,
                                                      Scalar aspect_ratio,
                                                      Scalar z_near,
                                                      Scalar z_far) -> frustum {
        return {.left   = -scale,
                .right  = scale,
                .bottom = -scale * aspect_ratio,
                .top    = scale * aspect_ratio,
                .z_near = z_near,
                .z_far  = z_far};
    }
};

} // namespace pw

#endif
